//STRESSBALANCE class definition
//
//   Usage:
//      stressbalance=new stressbalance();

function stressbalance (){
	//methods
	this.setdefaultparameters = function(){// {{{

		//maximum of non-linear iterations.
		this.maxiter=100;

		//Convergence criterion: absolute, relative and residual
		this.restol=Math.pow(10,-4); 
		this.reltol=0.01;
		this.abstol=10;

		this.FSreconditioning=Math.pow(10,13);
		this.shelf_dampening=0;

		//Penalty factor applied kappa=max(stiffness matrix)*10^penalty_factor
		this.penalty_factor=3;

		//coefficient to update the viscosity between each iteration of
		//a stressbalance according to the following formula
		//viscosity(n)=viscosity(n)+viscosity_overshoot(viscosity(n)-viscosity(n-1))
		this.viscosity_overshoot=0;

		//Stop the iterations of rift if below a threshold
		this.rift_penalty_threshold=0;

		//in some solutions, it might be needed to stop a run when only
		//a few constraints remain unstable. For thermal computation, this
		//parameter is often used.
		this.rift_penalty_lock=10;

		//output default:
		this.requested_outputs=['default'];

	}// }}}
	this.disp= function(){// {{{
		console.log(sprintf('   StressBalance solution parameters:'));

		console.log(sprintf('\n      %s','Convergence criteria:'));
		fielddisplay(this,'restol','mechanical equilibrium residual convergence criterion');
		fielddisplay(this,'reltol','velocity relative convergence criterion, NaN: not applied');
		fielddisplay(this,'abstol','velocity absolute convergence criterion, NaN: not applied');
		fielddisplay(this,'isnewton',"0: Picard's fixed point, 1: Newton's method, 2: hybrid");
		fielddisplay(this,'maxiter','maximum number of nonlinear iterations');
		fielddisplay(this,'viscosity_overshoot','over-shooting constant new=new+C*(new-old)');

		console.log(sprintf('\n      %s','boundary conditions:'));
		fielddisplay(this,'spcvx','x-axis velocity constraint (NaN means no constraint) [m/yr]');
		fielddisplay(this,'spcvy','y-axis velocity constraint (NaN means no constraint) [m/yr]');
		fielddisplay(this,'spcvz','z-axis velocity constraint (NaN means no constraint) [m/yr]');

		console.log(sprintf('\n      %s','Rift options:'));
		fielddisplay(this,'rift_penalty_threshold','threshold for instability of mechanical constraints');
		fielddisplay(this,'rift_penalty_lock','number of iterations before rift penalties are locked');

		console.log(sprintf('\n      %s','Penalty options:'));
		fielddisplay(this,'penalty_factor','offset used by penalties: penalty = Kmax*10^offset');
		fielddisplay(this,'vertex_pairing','pairs of vertices that are penalized');

		console.log(sprintf('\n      %s','Other:'));
		fielddisplay(this,'shelf_dampening','use dampening for floating ice ? Only for FS model');
		fielddisplay(this,'FSreconditioning','multiplier for incompressibility equation. Only for FS model');
		fielddisplay(this,'referential','local referential');
		fielddisplay(this,'loadingforce','loading force applied on each point [N/m^3]');
		fielddisplay(this,'requested_outputs','additional outputs requested');



	}// }}}
		this.checkconsistency = function(md,solution,analyses) { //{{{

			//Early return
			if(ArrayAnyEqual(ArrayIsMember(StressbalanceAnalysisEnum(),analyses),0))return;

			checkfield(md,'fieldname','stressbalance.spcvx','timeseries',1);
			checkfield(md,'fieldname','stressbalance.spcvy','timeseries',1);
			checkfield(md,'fieldname','stressbalance.spcvz','timeseries',1);
			checkfield(md,'fieldname','stressbalance.restol','size',[1, 1],'>',0,'NaN',1);
			checkfield(md,'fieldname','stressbalance.reltol','size',[1, 1]);
			checkfield(md,'fieldname','stressbalance.abstol','size',[1, 1]);
			checkfield(md,'fieldname','stressbalance.isnewton','numel',[1],'values',[0, 1, 2]);
			checkfield(md,'fieldname','stressbalance.FSreconditioning','size',[1, 1],'NaN',1);
			checkfield(md,'fieldname','stressbalance.viscosity_overshoot','size',[1, 1],'NaN',1);
			checkfield(md,'fieldname','stressbalance.maxiter','size',[1, 1],'>=',1);
			checkfield(md,'fieldname','stressbalance.referential','size',[md.mesh.numberofvertices, 6]);
			checkfield(md,'fieldname','stressbalance.loadingforce','size',[md.mesh.numberofvertices, 3]);
			checkfield(md,'fieldname','stressbalance.requested_outputs','stringrow',1);

			//singular solution
			if(!ArrayAnyNaN(md.stressbalance.spcvx) | !ArrayAnyNaN(md.stressbalance.spcvy) |  !ArrayAnyAboveStrict(md.mask.groundedice_levelset,0)){
				md = checkmessage(md,'model is not well posed (singular). You need at least one node with fixed velocity!');
				console.log(sprintf('\n !!! Warning: no spc applied, model might not be well posed if no basal friction is applied, check for solution crash\n'));
			}
			//CHECK THAT EACH LINES CONTAINS ONLY NAN VALUES OR NO NAN VALUES
			for(var i=0;i<md.stressbalance.referential.length;i++){
				var sum=0;
				for(j=0;j<md.stressbalance.referential[0].length;j++)sum+=md.stressbalance.referential[i][j];
				if (sum!=0 & sum!=6){
					md = checkmessage(md,'Each line of stressbalance.referential should contain either only NaN values or no NaN values');
					break;
				}
			}
			//CHECK THAT THE TWO VECTORS PROVIDED ARE ORTHOGONAL
			for(var i=0;i<md.stressbalance.referential.length;i++){
				var sum=0;
				for(j=0;j<md.stressbalance.referential[0].length;j++)sum+=md.stressbalance.referential[i][j];
				if(sum==0){
					var dot=0;
					for(j=0;j<3;j++)dot+=md.stressbalance.referential[i][j]*md.stressbalance.referential[i][j+3];
					dot=Math.abs(dot);
					if(dot>Math.pow(10,-18)){
						md.checkmessage('Vectors in stressbalance.referential (columns 1 to 3 and 4 to 6) must be orthogonal');
						break;
					}
				}
			}
			//CHECK THAT NO rotation specified for FS Grounded ice at base
			if (md.mesh.domaintype() == '3D' & md.flowequation.isFS){
				for(var i=0;i<md.mask.groundedice_levelset.length;i++){
					if(md.mask.groundedice_levelset[i]>0 & md.mesh.vertexonbase[i]){
						if(!ArrayIsNan(md.stressbalance.referential[i])){
							md.checkmessage('no referential should be specified for basal vertices of grounded ice');
							break;
						}
					}
				}
				checkfield(md,'fieldname','stressbalance.FSreconditioning','>',0);
			}
		} // }}}
	//properties 
	// {{{
	this.spcvx                  = NaN;
	this.spcvy                  = NaN;
	this.spcvz                  = NaN;
	this.restol                 = 0;
	this.reltol                 = 0;
	this.abstol                 = 0;
	this.isnewton               = 0;
	this.FSreconditioning       = 0;
	this.viscosity_overshoot    = 0;
	this.maxiter                = 0;
	this.shelf_dampening        = 0;
	this.vertex_pairing         = NaN;
	this.penalty_factor         = NaN;
	this.rift_penalty_lock      = NaN;
	this.rift_penalty_threshold = 0;
	this.referential            = NaN;
	this.loadingforce           = NaN;
	this.requested_outputs      = ['']

	this.setdefaultparameters();
	//}}}
}
